/*
 * $Id: restore.c,v 1.1.1.1 2002/05/11 21:39:59 mitch Exp $
 *
 * dumptool - A tool to manage MR-AFS dump files
 *
 * The dump file format ought to be documented _somewhere_, and
 * this seems like a good as a place as any ...
 *
 * A AFS dump file is marked off into a series of sections.  Each
 * section marked by a dump tag.  A tag is a single byte who's value
 * corresponds with the next section.  The sections are (in order):
 *
 * DUMPHEADER (tag 0x01)
 * VOLUMEHEADER (tag 0x02)
 * VNODE (tag 0x03)
 * DUMPEND (tag 0x04)
 *
 * Descriptions of the sections follow.  Note that in all cases, data is
 * stored in the dump in network byte order.
 *
 * DUMPHEADER:
 *
 * DUMPHEADER contains two parts: the DUMPMAGIC magic number (32 bits)
 * and the dump header itself.
 *
 * The dump header itself consists of a series of tagged values,
 * each tag marking out members of the DumpHeader structure.  The
 * routine ReadDumpHeader explains the specifics of these tags.
 *
 * VOLUMEHEADER:
 *
 * VOLUMEHEADER is a series of tagged values corresponding to the elements
 * of the VolumeDiskData structure.  See ReadVolumeHeader for more
 * information
 *
 * VNODE:
 *
 * The VNODE section is all vnodes contained in the volume (each vnode
 * itself is marked with the VNODE tag, so it's really a sequence of
 * VNODE tags, unlike other sections).
 *
 * Each vnode consists of three parts: the vnode number (32 bits), the
 * uniqifier (32 bits), and a tagged list of elements corresponding to
 * the elements of the VnodeDiskData structure.  See ScanVnodes for
 * more information.  Note that if file data is associated with a vnode,
 * it will be contained here.
 *
 * DUMPEND:
 *
 * The DUMPEND section consists of one part: the DUMPENDMAGIC magic
 * number (32 bits).
 * 
 * Notes:
 *
 * The tagged elements are all ASCII letters, as opposed to the section
 * headers (which are 0x01, 0x02, ...).  Thus, an easy way to tell when
 * you've reached the end of an element sequence is to check to see if
 * the next tag is a printable character (this code tests for < 20).
 *
 * "vos dump" dumps the large vnode index, then the small vnode index,
 * so directories will appear first in the VNODE section.
 */

#include <stdio.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/param.h>
#include <netinet/in.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <termios.h>
#include <fnmatch.h>

#include <lock.h>
#include <afs/param.h>
#include <afs/afsint.h>
#include <afs/nfs.h>
#include <afs/acl.h>
#if !defined(PRE_AFS_36) && !defined(RESIDENCY)
#include <afs/ihandle.h>
#endif /* !defined(PRE_AFS_36) && !defined(RESIDENCY) */
#include <afs/vnode.h>
#include <afs/volume.h>

#include <afs/dir.h>

#include "config.h"

/*
 * Sigh.  Linux blows it again
 */

#ifdef linux
#include <pty.h>
#endif

#ifdef LINT
RCSID("$Id: restore.c,v 1.1.1.1 2002/05/11 21:39:59 mitch Exp $")
#endif

/*
 * Stuff that is in private AFS header files, unfortunately
 */

#define DUMPVERSION	1
#define DUMPENDMAGIC	0x3A214B6E
#define DUMPBEGINMAGIC	0xB3A11322
#define D_DUMPHEADER	1
#define D_VOLUMEHEADER	2
#define D_VNODE		3
#define D_DUMPEND	4
#define D_MAX		20

#define MAXDUMPTIMES	50

   struct DumpHeader {
	 int32_t version;
	 VolumeId volumeId;
	 char volumeName[VNAMESIZE];
	 int nDumpTimes;             /* Number of pairs */
	 struct {
	       int32_t from, to;
	 } dumpTimes[MAXDUMPTIMES];
   };

/*
 * Our command-line arguments
 */

static int verbose = 0;
static int numNoDirData = 0;
static int termsize = 0;
int Testing = 0;

/*
 * We use this structure to hold vnode data in our hash table.
 * It's indexed by vnode number.
 */

struct vnodeData {
      struct VnodeDiskObject *vnode;	/* A pointer to the disk vnode */
      int vnodeNumber;		/* The vnode number */
      long dumpdata;			/* File offset of dump data (if
					   available */
      char *filename;                 /* File name */
      unsigned char *filedata;	/* A pointer to the actual file
				   data itself (if available) */
      unsigned int datalength;	/* The length of the data */
};

/*
 * This contains the current location when we're doing a scan of a
 * directory.
 */

struct DirCursor {
      int hashbucket;			/* Current hash bucket */
      int entry;			/* Entry within hash bucket */
};

/*
 * Arrays to hold vnode data
 */


typedef struct Node_val Node;

struct Node_val {
      struct vnodeData *data;
      Node *next;
};

void print(Node **List)
{
   Node *t;
   
   t = *List;
   while(t)
   {
      fprintf(stderr, "%s\n", t->data->filename);
      t = t->next;
   }
}

void insert(Node **List, struct vnodeData *vdata)
{
   Node *t;
   
   t = (Node *) malloc(sizeof(Node));
   t->data = vdata;
   t->next = *List;
   *List = t;

}


/* If i == -1, then chop off the last element of the list and return
   it, otherwise find the element of the list whose
   data->vnodeNumber == i, and return it without altering the list structure
*/

struct vnodeData *get(int i, Node **List)
{
   Node *L = *List;
   Node *t, *t1;
   Node *t2;
   struct vnodeData *vdata;
   
   if(L)
   {
      t = L;
      t1 = NULL;
      while(t->next && (t->data->vnodeNumber != i || i==-1))
      {
	 t1 = t;
	 t = t->next;
      }

      /* found the end of the list and had to chop it off */
      if(i == -1 && t->next == NULL)
      {
	 if(t1)
	 {
	    t1->next = NULL;
	 }
	 else
	 {
	    *List = NULL;
	 }
	 
	 vdata = t->data;
	 free(t);
	 return vdata;
      }
      /* found the element we were looking for */
      else if(t->data->vnodeNumber == i)
      {
	 return(t->data);
      }
   }
   return NULL;
}

/* returns 1 and deletes the node from the list if a matching node is
   found, otherwise simply returns 0 */

int search_delete(Node **List, char *key)
{
   Node *p, *q;
   p = *List;
   q = NULL;
      
   while(p)
   {
/*      if((strcmp(p->data->filename, key) == 0) ||
	 (p->data->filename[strlen(p->data->filename)-1] == '/' &&
	 (strncmp(p->data->filename, key, strlen(p->data->filename)-1) == 0))) */
      if(strcmp(p->data->filename, key) == 0)
      {
	 /* delete this node */
	 if(q)
	 {
	    q->next = p->next;
	 }
	 else
	 {
	    *List = p->next;
	 }
	 free(p->data->filename);
	 free(p->data);
	 free(p);

	 /* found */
	 return 1;
      }

      /* advance */
      q = p;
      p = p->next;
   }
   /* not found */
   return 0;
}

int search(Node **List, char *key, int n)
{
   Node *p, *q;
   p = *List;
   q = NULL;
      
   while(p)
   {
      if((strlen(p->data->filename) == n) && strncmp(key, p->data->filename, n) == 0)
      {
	 /* found */
	 return 1;
      }
      /* advance */
      q = p;
      p = p->next;
   }
   /* not found */
   return 0;
}


void clear(Node **List)
{
   Node *p;
   while(*List)
   {
      p = *List;
      *List = p->next;
      
      free(p->data->filename);
      if(p->data->filedata)
      {
	 free(p->data->filedata);
      }
      if(p->data->vnode)
      {
	 free(p->data->vnode);
      }
      free(p->data);
      free(p);
   }
}


/* !!!TODO!!!
  as is -- dirty hack. the way to do this is to implement linked
  lists, and add vnodes to them incrementing num{Large,Small}Vnodes,
  afterwards copy the contents of linked lists into
  {Large,Small}VnodeIndex and proceed as before. That way we avoid
  having to recode the qsort/bsearch, and other stuff.
*/

Node *VnodeList = NULL;
Node *ToExtract = NULL;
Node *DirToCreate = NULL;

struct vnodeData **LargeVnodeIndex;
struct vnodeData **SmallVnodeIndex;
int numLargeVnodes = 0;
int numSmallVnodes = 0;

/*
 * Crap for the libraries */

int ShutdownInProgress = 0;

/*
 * Our local function prototypes
 */

static int ReadDumpHeader(FILE *, struct DumpHeader *);
static int ReadVolumeHeader(FILE *, VolumeDiskData *);
static int ScanVnodes(FILE *, VolumeDiskData *, int);
static int DumpVnodeFile(FILE *, struct VnodeDiskObject *, VolumeDiskData *);
static struct vnodeData *InsertVnode(unsigned int, struct VnodeDiskObject *);
static struct vnodeData *GetVnode(unsigned int);
static int CompareVnode(const void *, const void *);
static void InteractiveRestore(FILE *, VolumeDiskData *);
static void DirectoryList(int, char **, struct vnodeData *, VolumeDiskData *);
static void DirListInternal(struct vnodeData *, char *[], int, int, int, int,
			    int, int, VolumeDiskData *, char *);
static int CompareDirEntry(const void *, const void *);
static struct vnodeData *ChangeDirectory(int, char **, struct vnodeData *);
static void CopyFile(int, char **, struct vnodeData *, FILE *);
static void CopyVnode(int, char **, FILE *);
static void DumpAllFiles(int, char **, struct vnodeData *, VolumeDiskData *);
static void DumpAllResidencies(FILE *, struct vnodeData *, VolumeDiskData *);
static struct vnodeData *FindFile(struct vnodeData *, char *);
static void ResetDirCursor(struct DirCursor *, struct vnodeData *);
static struct DirEntry *ReadNextDir(struct DirCursor *, struct vnodeData *);
static void MakeArgv(char *, int *, char ***);
static char *GetToken(char *, char **, char *, char *[]);
static int ReadInt16(FILE *, uint16_t *);
static int ReadInt32(FILE *, uint32_t *);
static int ReadString(FILE *, char *, int);
static int ReadByteString(FILE *, void *, int);

int
main(int argc, char *argv[])
{
   int c, errflg = 0, dumpvnodes = 0, force = 0, inode = 0;
   unsigned int magic;
   struct DumpHeader dheader;
   VolumeDiskData vol;
   long offset;
   int i, j;
   char *p;
   struct winsize win;
   struct vnodeData *vdata;
   FILE *f;
   int tag;

   if(argc==1)
   {
      fprintf(stderr, "%s\nRun this utility as follows:\n\t%s file_1 file_2 ...\n", AMANDA_AFS_VERSION, argv[0]);
      fprintf(stderr, "It will take its standard input, and, provided it is a vos dump\n");
      fprintf(stderr, "and files file_i, specified with full path relative to the dump root, exist\n");
      fprintf(stderr, "it will restore files file_i into the current directory.\n");
      return 0;
   }


   /* Get the list of files to restore into a linked list */
   i = 1;
   while(i<argc)
   {
      char *cur_name;
      char *t;
      int k, l;

      /* get rid of the trailing '/' */
      l = (argv[i][strlen(argv[i])-1] == '/') ? strlen(argv[i])-1 : strlen(argv[i]);

      /* is the entry w/o the slash in the exrtact list? */
      if(!search(&ToExtract, argv[i], l))
      {
	 /* No, thus add it */
	 vdata = (struct vnodeData *) malloc(sizeof(struct vnodeData));
	 vdata->vnode = NULL;
	 vdata->vnodeNumber = 0;
	 vdata->dumpdata = 0;
	 vdata->filedata = NULL;
	 vdata->datalength = 0;
	 vdata->filename = (char *) malloc(sizeof(char)*(l+1));
//	 t = (char *) malloc(l*sizeof(char));
	 k = 0;
	 memcpy(vdata->filename, argv[i], l);
	 memcpy(vdata->filename+l, &k, 1);
//	 sprintf(vdata->filename, "%s", t);
//	 free(t);
	 insert(&ToExtract, vdata);
      }

      /* Yes, it is in the extract list in that case, we must check
         whether any of its ancestors are in the extract list or
         whether all of the ancestors are in the create list */
      
      for(k = 2; k <= l; k++)
      {
	 if((argv[i])[k] == '/')
	 {
	    /* found an ancestor */
	    if(!search(&ToExtract, argv[i], k))
	    {
	       /* ancestor is not to be extracted */
	       if(!search(&DirToCreate, argv[i], k))
	       {
		  /* ancestor is not yet to be created, so create it */
		  vdata = (struct vnodeData *) malloc(sizeof(struct vnodeData));
		  vdata->vnode = NULL;
		  vdata->vnodeNumber = 0;
		  vdata->dumpdata = 0;
		  vdata->filedata = NULL;
		  vdata->datalength = 0;
		  vdata->filename = (char *) malloc(sizeof(char)*(k+1));
		  t = (char *) malloc(l);
		  memcpy(t, argv[i], k);
		  sprintf(vdata->filename, "%s\000", t);
		  free(t);
		  insert(&DirToCreate, vdata);
	       }
	       else
	       {
		  /* ancrestor is to be created, but must check further down to guarantee this holds */
		  continue;
	       }
	    }
	    else
	    {
	       /* this ancestor is to be extracted completely, so no need to search down */
	       break;
	    }
	 }
      }

      /* end of ancestor scouting */
      i++;
   }

   /* This is how we plan to do it (as opposed to "This is how we do it!"):
      1. Read in the command line arguments specifying the things we need to restore. (done)
      2. Open the input file (stdin) and parse the archive header.
      3. Start parsing the vnode list.
      3.a. Check vnode->parent, parse its structure
      (it is a directory) and find the filename that corresponds
      to the vnodeNumber.
      3.b. If that filename (with full path within the archive)
      matches anything on the list, then restore this file
      (creating the directory tree)
      3.c. If that filename on the ToExtract list is a directory,
      then create it, and add all of its contents to the
      ToExtract list.

      Keep doing this until the ToExtract list is empty
   */
	   
   /* Try opening the dump file */

   f = stdin;

   while(!feof(f))
   {
      if ((tag = getc(f)) == D_DUMPHEADER)
      {
	 ungetc(tag, f);
      }
      else
      {
	 break;
      }
      
      if (ReadDumpHeader(f, &dheader)) {
	 fprintf(stderr, "Failed to read dump header!\n");
	 exit(1);
      }
      
      if (verbose)
	 printf("Dump is for volume %lu (%s)\n", dheader.volumeId,
		dheader.volumeName);
      
      if (getc(f) != D_VOLUMEHEADER) {
	 fprintf(stderr, "Volume header is missing from dump, aborting\n");
	 exit(1);
      }
      
      if (ReadVolumeHeader(f, &vol)) {
	 fprintf(stderr, "Unable to read volume header\n");
	 exit(1);
      }
      
      if (verbose) {
	 printf("Volume information:\n");
	 printf("\tid = %lu\n", vol.id);
	 printf("\tparent id = %lu\n", vol.parentId);
	 printf("\tname = %s\n", vol.name);
	 printf("\tflags =");
	 if (vol.inUse)
	    printf(" inUse");
	 if (vol.inService)
	    printf(" inService");
	 if (vol.blessed)
	    printf(" blessed");
	 if (vol.needsSalvaged)
	    printf(" needsSalvaged");
	 printf("\n");
	 printf("\tuniquifier = %lu\n", vol.uniquifier);
	 printf("\tCreation date = %s", ctime((time_t *) &vol.creationDate));
	 printf("\tLast access date = %s", ctime((time_t *) &vol.accessDate));
	 printf("\tLast update date = %s", ctime((time_t *) &vol.updateDate));
	 printf("\tVolume owner = %lu\n", vol.owner);
      }
      
      if (verbose)
	 printf("Scanning vnodes (this may take a while)\n");
      
      /*
       * We need to do two vnode scans; one to get the number of
       * vnodes, the other to actually build the index.
       */

      /* most modifications will take place in ScanVnodes */
      
      if (ScanVnodes(f, &vol, 0)) {
	 fprintf(stderr, "Second vnode scan failed, aborting\n");
	 exit(1);
      }

      if (getc(f) != D_DUMPEND || ReadInt32(f, (uint32_t *)&magic) ||
	  magic != DUMPENDMAGIC) {
	 fprintf(stderr, "Couldn't find dump postamble, ");
	 if (! force) {
	    fprintf(stderr, "aborting (use -f to override)\n");
	    exit(1);
	 } else {
	    fprintf(stderr, "continuing anyway\n");
	    fprintf(stderr, "WARNING: Dump may not be complete!\n");
	 }
      }
      
      /*
       * If we wanted to simply dump all vnodes, do it now
       */

      /* now to make sure we're not leaking, clear VnodeList. */
      clear(&VnodeList);
   }

   return 0;
}

/*
 * Read the dump header, which is at the beginning of every dump
 */

static int
ReadDumpHeader(FILE *f, struct DumpHeader *header)
{
   unsigned int magic;
   int tag, i;

   if (getc(f) != D_DUMPHEADER ||
       ReadInt32(f, &magic) || ReadInt32(f, (uint32_t *)(unsigned int *)
					 &header->version) ||
       magic != DUMPBEGINMAGIC) {
      if (verbose)
	 fprintf(stderr, "Couldn't find dump magic numbers\n");
      return -1;
   }

   header->volumeId = 0;
   header->nDumpTimes = 0;

   while ((tag = getc(f)) > D_MAX && tag != EOF) {
      unsigned short length;
      switch (tag) {
	 case 'v':
	    if (ReadInt32(f, (uint32_t *)&header->volumeId)) {
	       if (verbose)
		  fprintf(stderr, "Failed to read "
			  "volumeId\n");
	       return -1;
	    }
	    break;
	 case 'n':
	    if (ReadString(f, header->volumeName,
			   sizeof(header->volumeName))) {
	       if (verbose)
		  fprintf(stderr, "Failed to read "
			  "volume name\n");
	       return -1;
	    }
	    break;
	 case 't':
	    if (ReadInt16(f, &length)) {
	       if (verbose)
		  fprintf(stderr, "Failed to read "
			  "dump time array length\n");
	       return -1;
	    }
	    header->nDumpTimes = (length >> 1);
	    for (i = 0; i < header->nDumpTimes; i++)
	       if (ReadInt32(f, (unsigned int *)
			     &header->dumpTimes[i].from) ||
		   ReadInt32(f, (unsigned int *)
			     &header->dumpTimes[i].to)) {
		  if (verbose)
		     fprintf(stderr, "Failed to "
			     "read dump times\n");
		  return -1;
	       }
	    break;
	 default:
	    if (verbose)
	       fprintf(stderr, "Unknown dump tag \"%c\"\n",
		       tag);
	    return -1;
      }
   }

   if (!header->volumeId || !header->nDumpTimes) {
      if (verbose) 
	 fprintf(stderr, "We didn't get a volume Id or "
		 "dump times listing\n");
      return 1;
   }

   ungetc(tag, f);
   return 0;
}

/*
 * Read the volume header; this is the information stored in VolumeDiskData.
 *
 * I'm not sure we need all of this, but read it in just in case.
 */

static int
ReadVolumeHeader(FILE *f, VolumeDiskData *vol)
{
   int tag;
   unsigned int trash;
   memset((void *) vol, 0, sizeof(*vol));

   while ((tag = getc(f)) > D_MAX && tag != EOF) {
      switch (tag) {
	 case 'i':
	    if (ReadInt32(f, (uint32_t *)&vol->id))
	       return -1;
	    break;
	 case 'v':
	    if (ReadInt32(f, (uint32_t *)&trash))
	       return -1;
	    break;
	 case 'n':
	    if (ReadString(f, vol->name, sizeof(vol->name)))
	       return -1;
	    break;
	 case 's':
	    vol->inService = getc(f);
	    break;
	 case 'b':
	    vol->blessed = getc(f);
	    break;
	 case 'u':
	    if (ReadInt32(f, (uint32_t *)&vol->uniquifier))
	       return -1;
	    break;
	 case 't':
	    vol->type = getc(f);
	    break;
	 case 'p':
	    if (ReadInt32(f, (uint32_t *)&vol->parentId))
	       return -1;
	    break;
	 case 'c':
	    if (ReadInt32(f, (uint32_t *)&vol->cloneId))
	       return -1;
	    break;
	 case 'q':
	    if (ReadInt32(f, (uint32_t *) &vol->maxquota))
	       return -1;
	    break;
	 case 'm':
	    if (ReadInt32(f, (uint32_t *) &vol->minquota))
	       return -1;
	    break;
	 case 'd':
	    if (ReadInt32(f, (uint32_t *) &vol->diskused))
	       return -1;
	    break;
	 case 'f':
	    if (ReadInt32(f, (uint32_t *) &vol->filecount))
	       return -1;
	    break;
	 case 'a':
	    if (ReadInt32(f, (uint32_t *)&vol->accountNumber))
	       return -1;
	    break;
	 case 'o':
	    if (ReadInt32(f, (uint32_t *)&vol->owner))
	       return -1;
	    break;
	 case 'C':
	    if (ReadInt32(f, (uint32_t *)&vol->creationDate))
	       return -1;
	    break;
	 case 'A':
	    if (ReadInt32(f, (uint32_t *)&vol->accessDate))
	       return -1;
	    break;
	 case 'U':
	    if (ReadInt32(f, (uint32_t *)&vol->updateDate))
	       return -1;
	    break;
	 case 'E':
	    if (ReadInt32(f, (uint32_t *)&vol->expirationDate))
	       return -1;
	    break;
	 case 'B':
	    if (ReadInt32(f, (uint32_t *)&vol->backupDate))
	       return -1;
	    break;
	 case 'O':
	    if (ReadString(f, vol->offlineMessage,
			   sizeof(vol->offlineMessage)))
	       return -1;
	    break;
	 case 'M':
	    if (ReadString(f, (char *) vol->stat_reads, VMSGSIZE))
	       return -1;
	    break;
	 case 'W': {
	    unsigned short length;
	    int i;
	    unsigned int data;
	    if (ReadInt16(f, &length))
	       return -1;
	    for (i = 0; i < length; i++) {
	       if (ReadInt32(f, (uint32_t *)&data))
		  return -1;
	       if (i < sizeof(vol->weekUse) /
		   sizeof(vol->weekUse[0]))
		  vol->weekUse[i] = data;
	    }
	    break;
	 }
	 case 'D':
	    if (ReadInt32(f, (uint32_t *)&vol->dayUseDate))
	       return -1;
	    break;
	 case 'Z':
	    if (ReadInt32(f, (uint32_t *) &vol->dayUse))
	       return -1;
	    break;
	 default:
	    if (verbose)
	       fprintf(stderr, "Unknown dump tag \"%c\"\n",
		       tag);
	    return -1;
      }
   }

   ungetc(tag, f);
   return 0;
}

/*
 * Scan all our vnode entries, and build indexing information.
 */

static int
ScanVnodes(FILE *f, VolumeDiskData *vol, int sizescan)
{
   int vnodeNumber;
   int tag;
   int numFileVnodes = 0;
   int numDirVnodes = 0;
   unsigned char buf[SIZEOF_LARGEDISKVNODE];
   struct VnodeDiskObject *vnode = (struct VnodeDiskObject *) buf;
   struct VnodeDiskObject *nvnode;
   long offset, oldoffset;
   struct vnodeData *vdata;
   unsigned int length;
   char *iobuf;
   int i;
   int error = 0;
   int out;
   int extract, create;

   tag = getc(f);

   while (tag == D_VNODE) {

      offset = 0;
      length = 0;
      vnode->type = -1;
      vnode->length = -1;

      if (ReadInt32(f, (uint32_t *) &vnodeNumber))
      {
	 fprintf(stderr, "failed int32 for 'vnodenum'\n");
	 return -1;
      }

      if (ReadInt32(f, (uint32_t *)&vnode->uniquifier))
      {
	 fprintf(stderr, "failed int32 for 'uniquifier'\n");
	 return -1;
      }
		
      if (verbose > 1 && !sizescan)
	 printf("Got vnode %d\n", vnodeNumber);
		
      while ((tag = getc(f)) > D_MAX && tag != EOF)
	 switch (tag) {
	    case 't':
	       vnode->type = (VnodeType) getc(f);
	       break;
	    case 'l':
	    {
	       unsigned short tmp;
	       if (ReadInt16(f, &tmp))
	       {
		  fprintf(stderr, "failed int16 for 'l'\n");
		  return -1;
	       }
	       vnode->linkCount = tmp;
	    }
	    break;
	    case 'v':
	       if (ReadInt32(f, (uint32_t *)&vnode->dataVersion))
	       {
		  fprintf(stderr, "failed int32 for 'v'\n");
		  return -1;
	       }
	       break;
	    case 'm':
	       if (ReadInt32(f, (uint32_t *) &vnode->unixModifyTime))
	       {
		  fprintf(stderr, "failed int32 for 'm'\n");
		  return -1;
	       }
	       break;
	    case 's':
	       if (ReadInt32(f, (uint32_t *) &vnode->serverModifyTime))
	       {
		  fprintf(stderr, "failed int32 for 's'\n");
		  return -1;
	       }
	       break;
	    case 'a':
	       if (ReadInt32(f, (uint32_t *)&vnode->author))
	       {
		  fprintf(stderr, "failed int32 for 'a'\n");
		  return -1;
	       }
	       break;
	    case 'o':
	       if (ReadInt32(f, (uint32_t *)&vnode->owner))
	       {
		  fprintf(stderr, "failed int32 for 'o'\n");
		  return -1;
	       }
	       break;
	    case 'g':
	       if (ReadInt32(f, (uint32_t *) &vnode->group))
	       {
		  fprintf(stderr, "failed int32 for 'g'\n");
		  return -1;
	       }
	       break;
	    case 'b': {
	       unsigned short modeBits;
	       if (ReadInt16(f, &modeBits))
		  return -1;
	       vnode->modeBits = modeBits;
	       break;
	    }
	    case 'p':
	       if (ReadInt32(f, (uint32_t *)&vnode->parent))
	       {
		  fprintf(stderr, "failed int32 for 'p'\n");
		  return -1;
	       }
	       break;
	    case 'S':
	       if (ReadInt32(f, (uint32_t *)&vnode->length))
	       {
		  fprintf(stderr, "failed int32 for 'S'\n");
		  return -1;
	       }
	       break;
	    case 'F':
	       if (ReadInt32(f, (uint32_t *) &vnode->vn_ino_lo))
		  return -1;
	       break;
	    case 'A':
	       if (ReadByteString(f,
				  (void *)VVnodeDiskACL(vnode),
				  VAclDiskSize(vnode)))
	       {
		  fprintf(stderr, "failed readbystring for 'A'\n");
		  return -1;
	       }
#if 0
	       acl_NtohACL(VVnodeDiskACL(vnode));
#endif
	       break;
	    case 'f':
	       if (verbose > 1 && ! sizescan)
		  printf("We have file data!\n");
	       if (ReadInt32(f, (uint32_t *)&length))
	       {
		  fprintf(stderr, "failed int32 for 'f'\n");
		  return -1;
	       }
	       vnode->length = length;
/* stuff */

	       if (vnode->type == -1)
		  break;
			      
/*	       if (vnode->type == vDirectory)
		  numDirVnodes++;
	       else
		  numFileVnodes++;
*/

	       nvnode = (struct VnodeDiskObject *) malloc(sizeof(struct VnodeDiskObject));
			      
	       if (!nvnode) {
		  if (verbose)
		     fprintf(stderr, "Unable to allocate space for vnode\n");
		  return -1;
	       }
	       memcpy((void *) nvnode, (void *) vnode, sizeof(struct VnodeDiskObject));
			      
	       vdata = (struct vnodeData *) malloc(sizeof(struct vnodeData));
			      
	       vdata->vnode = nvnode;
	       vdata->vnodeNumber = vnodeNumber;
	       vdata->filedata = NULL;
	       vdata->datalength = length;
			      
	       insert(&VnodeList, vdata);
			      
	       if (vdata == NULL) {
		  if (verbose)
		     fprintf(stderr, "Failed to insert "
			     "vnode into hash table");
		  return -1;
	       }
	       
	       /*
		* Save directory data, since we'll need it later.
		*/
	       
	       if (vnode->type == vDirectory && length) {
		  
		  vdata->filedata = malloc(length);
				 
		  if (!vdata->filedata) {
		     if (verbose)
			fprintf(stderr, "Unable to "
				"allocate space for "
				"file data (%d)\n",
				length);
		     return -1;
		  }
		  
		  if (fread(vdata->filedata, length, 1, f) != 1) {
		     if (verbose)
			fprintf(stderr, "Unable to "
				"read in file data!\n");
		     return -1;
		  }

		  /* dealing with a directory */

		  if(vnode->parent != 0) 
		  {
		     /* dealing with a subdirectory of the root */

		     struct vnodeData *v;
		     struct DirEntry *ep;
		     struct DirCursor cursor;
		     
		     v = get(vnode->parent, &VnodeList);

		     /* find out the filename of the vnode we're dealing with, put it into f */

		     ResetDirCursor(&cursor, v);
		     
		     /*
		      * Scan through the whole directory
		      */
		     
		     while ((ep = ReadNextDir(&cursor, v)) != NULL)
		     {
			if (ntohl(ep->fid.vnode) == vdata->vnodeNumber && ntohl(ep->fid.vunique) == vdata->vnode->uniquifier)
			{
			   /* found our vnode */
			   vdata->filename = (char *) malloc((strlen(ep->name) + strlen(v->filename) + 2)*sizeof(char));
			   sprintf(vdata->filename, "%s/%s\000", v->filename, ep->name);
			   break;
			}
		     }
		  }
		  else /* in the root directory */
		  {
		     /* need to tell weather dealing with a volset or a single volume:

			for a volset the default case is to have vnode 1 equivalent to
		        './vol->name/'
			whereas for a single volume vnode 1 corresponds to ./

			can do that with command line options, need to modify gtar-wrapper.pl 
		     */

		     vdata->filename = (char *)malloc ((strlen(vol->name) + 3)*sizeof(char));
		     sprintf(vdata->filename, "./%s\000", vol->name);
		  }

		  /*
		    check to ToExtract list, if vdata->filename is on
		    it, then create it, and add all of its files onto
		    ToExtract list
		  */

		  extract = search_delete(&ToExtract, vdata->filename);
		  create = search_delete(&DirToCreate, vdata->filename);
		  
		  if(extract || create)
		  {
		     /* create the directory */
		     if(mkdir(vdata->filename, vdata->vnode->modeBits) == -1)
		     {
			if(errno == EEXIST)
			{
			   struct stat buf;
			   
			   stat(vdata->filename, &buf);
			   
			   if(!S_ISDIR(buf.st_mode))
			   {
			      error = 1;
			      fprintf(stderr, "Error: could not create directory!");
			   }
			}
		     }
		  }
		  if(!error && extract)
		  {
		     /* add all of its contents to the ToExtract list */
		     struct vnodeData *v, *tv, *new_v;
		     struct DirEntry *ep;
		     struct DirCursor cursor;
		     
		     v = get(vdata->vnodeNumber, &VnodeList);
		     
		     /* find out the filename of the vnode we're dealing with, put it into f */
		     
		     ResetDirCursor(&cursor, v);
		     
		     /*
		      * Scan through the whole directory
		      */
		     
		     while ((ep = ReadNextDir(&cursor, v)) != NULL)
		     {
			
			if(!strcmp(ep->name, ".") || !strcmp(ep->name, ".."))
			{
			   continue;
			}
			
			new_v = (struct vnodeData *) malloc(sizeof(struct vnodeData));
			/* here we should decide between whether the entry we're dealing with
			   is a directory or a file. If it is a directory it should have a '/'
			   at the end of its ToExtract entry, if it's not then it should have nothing extra.
			   
			   The problem is that at any given moment the vnode for ntohl(ep->fid.vnode)
			   might not have been encountered yet, and so data pretaining to it is inaccessible.
			   Thus, what needs to be done is search_delete() has to be modified.
			   
			*/
			new_v->vnode = NULL;
			new_v->vnodeNumber = 0;
			new_v->dumpdata = 0;
			new_v->filedata = NULL;
			new_v->datalength = 0;
			new_v->filename = (char *) malloc((strlen(ep->name)+strlen(v->filename)+2)*sizeof(char));
			sprintf(new_v->filename, "%s/%s\000", v->filename, ep->name);
			
			insert(&ToExtract, new_v);
		     }
		  }
	       } else if (vnode->type == vDirectory) {
		  /*
		   * Warn the user we may not have all directory
		   * vnodes
		   */
		  numNoDirData++;
	       } else if (length || (vnode->type != vDirectory && !length)) {

		  /* dealing with a regular file */

		  if(vnode->parent != 0) 
		  {
		     struct vnodeData *v;
		     struct DirEntry *ep;
		     struct DirCursor cursor;
		     
		     v = get(vnode->parent, &VnodeList);

		     /* find out the filename of the vnode we're dealing with, put it into f */

		     ResetDirCursor(&cursor, v);
		     
		     /*
		      * Scan through the whole directory
		      */
		     
		     while ((ep = ReadNextDir(&cursor, v)) != NULL)
		     {
			if (ntohl(ep->fid.vnode) == vdata->vnodeNumber && ntohl(ep->fid.vunique) == vdata->vnode->uniquifier)
			{
			   /* found our vnode */
			   vdata->filename = (char *) malloc((strlen(ep->name) + strlen(v->filename) + 2)*sizeof(char));
			   sprintf(vdata->filename, "%s/%s\000", v->filename, ep->name);
			   break;
			}
		     }
		  }
		  
		  if(search_delete(&ToExtract, vdata->filename))
		  {
		     /* create the file*/
		     int code;
		     out = open(vdata->filename, O_WRONLY | O_CREAT | O_TRUNC, vdata->vnode->modeBits);

		     i = length;
		     iobuf = (char *) malloc(BUF_SIZE*sizeof(char));
		     while(i > BUF_SIZE)
		     {
			code = fread(iobuf, 1, BUF_SIZE, f);
			if(code != BUF_SIZE)
			{
			   fprintf(stderr, "Errror: didn't read as much as was hoped for!\n");
			}
			i -= BUF_SIZE;
			code = write(out, iobuf, BUF_SIZE);
			if(code != BUF_SIZE)
			{
			   fprintf(stderr, "Error: didn't write as much as was hoped for!\n");
			}
		     }
		     code = fread(iobuf, 1, i, f);
		     if(code != i)
		     {
			fprintf(stderr, "Errror: didn't read as much as was hoped for!\n");
		     }
		     code = write(out, iobuf, i);
		     if(code != i)
		     {
			fprintf(stderr, "Error: didn't write as much as was hoped for!\n");
		     }
		     close(out);
		     free(iobuf);
		     
		  }
		  else
		  {
		     /* if we do not need this file, then skip its data in the input stream */
		     /* instead of fseek */
		     i = length;
		     iobuf = (char *) malloc(BUF_SIZE*sizeof(char));
		     while(i > BUF_SIZE)
		     {
			fread(iobuf, BUF_SIZE, 1, f);
			i -= BUF_SIZE;
		     }
		     fread(iobuf, i, 1, f);
		     free(iobuf);
		     /* end instead of fseek */
		  }
	       }

/* end stuff */
			
	       break;
	    default:
	       if (verbose)
		  fprintf(stderr, "Unknown dump tag \"%c\"\n",
			  tag);
	       return -1;
	 }
   }
	
	
   ungetc(tag, f);

   if (verbose)
      fprintf(stderr,"%s vnode scan completed\n",
	      sizescan ? "Primary" : "Secondary");

   return 0;
}

/*
 * Perform an interactive restore
 *
 * Parsing the directory information is a pain, but other than that
 * we just use the other tools we already have in here.
 */

static void
InteractiveRestore(FILE *f, VolumeDiskData *vol)
{
   struct vnodeData *vdatacwd;	/* Vnode data for our current dir */
   char cmdbuf[256];
   int argc;
   char **argv;

   /*
    * Let's see if we can at least get the data for our root directory.
    * If we can't, there's no way we can do an interactive restore.
    */

   if ((vdatacwd = GetVnode(1)) == NULL) {
      fprintf(stderr, "No entry for our root vnode!  Aborting\n");
      return;
   }

   if (! vdatacwd->filedata) {
      fprintf(stderr, "There is no directory data for the root "
	      "vnode (1.1).  An interactive\nrestore is not "
	      "possible.\n");
      return;
   }

   /*
    * If you're doing a selective dump correctly, then you should get all
    * directory vnode data.  But just in case you didn't, let the user
    * know there may be a problem.
    */

   if (numNoDirData)
      fprintf(stderr, "WARNING: %d directory vnodes had no file "
	      "data.  An interactive restore\nmay not be possible\n",
	      numNoDirData);

   printf("> ");
   while (fgets(cmdbuf, 256, stdin)) {

      cmdbuf[strlen(cmdbuf) - 1] = '\0';

      if (strlen(cmdbuf) == 0) {
	 printf("> ");
	 continue;
      }

      MakeArgv(cmdbuf, &argc, &argv);

      if (strcmp(argv[0], "ls") == 0) {
	 DirectoryList(argc, argv, vdatacwd, vol);
      } else if (strcmp(argv[0], "cd") == 0) {
	 struct vnodeData *newvdata;

	 newvdata = ChangeDirectory(argc, argv, vdatacwd);

	 if (newvdata)
	    vdatacwd = newvdata;
      } else if (strcmp(argv[0], "file") == 0) {
	 DumpAllFiles(argc, argv, vdatacwd, vol);
      } else if (strcmp(argv[0], "cp") == 0) {
	 CopyFile(argc, argv, vdatacwd, f);
      } else if (strcmp(argv[0], "vcp") == 0) {
	 CopyVnode(argc, argv, f);
      } else if (strcmp(argv[0], "quit") == 0 ||
		 strcmp(argv[0], "exit") == 0)
	 break;
      else if (strcmp(argv[0], "?") == 0 ||
	       strcmp(argv[0], "help") == 0) {
	 printf("Valid commands are:\n");
	 printf("\tls\t\tList current directory\n");
	 printf("\tcd\t\tChange current directory\n");
	 printf("\tcp\t\tCopy file from dump\n");
	 printf("\tvcp\t\tCopy file from dump (via vnode)\n");
	 printf("\tquit | exit\tExit program\n");
	 printf("\thelp | ?\tBrief help\n");
      } else 
	 fprintf(stderr, "Unknown command, \"%s\", enter "
		 "\"help\" for a list of commands.\n",
		 argv[0]);
		
      printf("> ");
   }

   return;
}

/*
 * Do a listing of all files in a directory.  Sigh, I wish this wasn't
 * so complicated.
 *
 * With the reorganizing, this is just a front-end to DirListInternal()
 */

static void
DirectoryList(int argc, char **argv, struct vnodeData *vdata,
	      VolumeDiskData *vol)
{
   int errflg = 0, lflag = 0, iflag = 0, Fflag = 0, sflag = 0, Rflag = 0;
   int c;

   optind = 1;

   while ((c = getopt(argc, argv, "liFRs")) != EOF)
      switch (c) {
	 case 'l':
	    lflag++;
	    break;
	 case 'i':
	    iflag++;
	    break;
	 case 'F':
	    Fflag++;
	    break;
	 case 'R':
	    Rflag++;
	 case 's':
	    sflag++;
	    break;
	 case '?':
	 default:
	    errflg++;
      }

   if (errflg) {
      fprintf(stderr, "Usage: %s [-liFs] filename [filename ...]\n",
	      argv[0]);
      return;
   }

   DirListInternal(vdata, &(argv[optind]), argc - optind, lflag, iflag,
		   Fflag, Rflag, 1, vol, NULL);

   return;
}

/*
 * Function that does the REAL work in terms of directory listing
 */

static void
DirListInternal(struct vnodeData *vdata, char *pathnames[], int numpathnames,
		int lflag, int iflag, int Fflag, int Rflag, int verbose,
		VolumeDiskData *vol, char *path)
{
   struct DirEntry *ep, **eplist = NULL, **eprecurse = NULL;
   struct DirCursor cursor;
   struct vnodeData *lvdata;

   int i, j, numentries = 0, longestname = 0, numcols, col, numrows;
   int numrecurse = 0;
		
   if (! vdata->filedata) {
      fprintf(stderr, "There is no vnode data for this "
	      "directory!\n");
      return;
   }

   ResetDirCursor(&cursor, vdata);

   /*
    * Scan through the whole directory
    */

   while ((ep = ReadNextDir(&cursor, vdata)) != NULL) {

      /*
       * If we didn't get any filenames on the command line,
       * get them all.
       */

      if (numpathnames == 0) {
	 eplist = realloc(eplist, sizeof(struct DirEntry *) *
			  ++numentries);
	 eplist[numentries - 1] = ep;
	 if (strlen(ep->name) > longestname)
	    longestname = strlen(ep->name);
	 if (Rflag)
	    if ((lvdata = GetVnode(ntohl(ep->fid.vnode))) &&
		lvdata->vnode->type == vDirectory &&
		!(strcmp(ep->name, ".") == 0 ||
		  strcmp(ep->name, "..") == 0)) {
	       eprecurse = realloc(eprecurse,
				   sizeof(struct DirEntry *) *
				   ++numrecurse);
	       eprecurse[numrecurse - 1] = ep;
	    }

      } else {
	 /*
	  * Do glob matching via fnmatch()
	  */

	 for (i = 0; i < numpathnames; i++)
	    if (fnmatch(pathnames[i], ep->name,
			FNM_PATHNAME) == 0) {
	       eplist = realloc(eplist,
				sizeof(struct DirEntry *) *
				++numentries);
	       eplist[numentries - 1] = ep;
	       if (strlen(ep->name) > longestname)
		  longestname = strlen(ep->name);
	       if (Rflag)
		  if ((lvdata =
		       GetVnode(ntohl(ep->fid.vnode))) &&
		      lvdata->vnode->type ==
		      vDirectory &&
		      !(strcmp(ep->name, ".") == 0 ||
			strcmp(ep->name, "..") == 0)) {
		     eprecurse =
			realloc(eprecurse,
				sizeof(struct DirEntry *) *
				++numrecurse);
		     eprecurse[numrecurse - 1] = ep;
		  }
	       break;
	    }
      }
   }

   qsort((void *) eplist, numentries, sizeof(struct DirEntry *),
	 CompareDirEntry);

   if (Rflag && eprecurse)
      qsort((void *) eprecurse, numrecurse,
	    sizeof(struct DirEntry *), CompareDirEntry);
   /*
    * We don't have to do column printing if we have the -l or the -i
    * options.  Sigh, column printing is WAY TOO FUCKING COMPLICATED!
    */

   if (!lflag && !iflag) {
      char c;

      if (Fflag)
	 longestname++;

      longestname++;

      numcols = termsize / longestname ? termsize / longestname : 1;
      numrows = numentries / numcols +
	 (numentries % numcols ? 1 : 0);

      for (i = 0; i < numrows; i++) {
	 col = 0;
	 while (col < numcols && (i + col * numrows) <
		numentries) {
	    ep = eplist[i + col++ * numrows];
	    if (Fflag) {
	       if (!(lvdata =
		     GetVnode(ntohl(ep->fid.vnode))))
		  c = ' ';
	       else if (lvdata->vnode->type ==
			vDirectory)
		  c = '/';
	       else if (lvdata->vnode->type ==
			vSymlink)
		  c = '@';
	       else if (lvdata->vnode->modeBits &
			0111 != 0)
		  c = '*';
	       else
		  c = ' ';
	       printf("%s%-*c", ep->name,
		      longestname - strlen(ep->name), c);
	    } else
	       printf("%-*s", longestname, ep->name);
	 }

	 printf("\n");
      }
   } else if (iflag)
      for (i = 0; i < numentries; i++)
	 if (!(lvdata = GetVnode(ntohl(eplist[i]->fid.vnode))))
	    printf("%d.0.0\t%s\n",
		   vol->parentId ? vol->parentId : vol->id,
		   eplist[i]->name);
	 else
	    if (path)
	       printf("%d.%d.%d\t%s/%s\n",
		      vol->id,
		      ntohl(eplist[i]->fid.vnode),
		      ntohl(eplist[i]->fid.vunique),
		      path, eplist[i]->name);
	    else
	       printf("%d.%d.%d\t%s\n",
		      vol->id,
		      ntohl(eplist[i]->fid.vnode),
		      ntohl(eplist[i]->fid.vunique),
		      eplist[i]->name);
   else if (lflag) {
      for (i = 0; i < numentries; i++)
	 if (!(lvdata = GetVnode(ntohl(eplist[i]->fid.vnode))))
	    printf("----------   0 0        "
		   "0                 0 %s\n",
		   eplist[i]->name);
	 else {
	    switch (lvdata->vnode->type) {
	       case vDirectory:
		  printf("d");
		  break;
	       case vSymlink:
		  printf("l");
		  break;
	       default:
		  printf("-");
	    }

	    for (j = 8; j >= 0; j--) {
	       if (lvdata->vnode->modeBits & (1 << j))
		  switch (j % 3) {
		     case 2: printf("r");
			break;
		     case 1: printf("w");
			break;
		     case 0: printf("x");
		  }
	       else
		  printf("-");
	    }

	    printf(" %-3d %-8d %-8d %10d %s\n",
		   lvdata->vnode->linkCount,
		   lvdata->vnode->owner,
		   lvdata->vnode->group,
		   lvdata->vnode->length,
		   eplist[i]->name);
	 }
   }

   free(eplist);

   if (Rflag && eprecurse) {
      char *lpath;
      lpath = NULL;
      for (i = 0; i < numrecurse; i++) {
	 if (verbose)
	    printf("\n%s:\n", eprecurse[i]->name);
	 if (path) {
	    lpath = malloc(strlen(path) +
			   strlen(eprecurse[i]->name) + 2);
	    if (lpath)
	       sprintf(lpath, "%s/%s", path,
		       eprecurse[i]->name);
	 }
	 DirListInternal(
	    GetVnode(ntohl(eprecurse[i]->fid.vnode)),
	    NULL, 0, lflag, iflag, Fflag, Rflag,
	    verbose, vol, lpath);
	 if (lpath) {
	    free(lpath);
	    lpath = NULL;
	 }
      }
   }

   if (eprecurse)
      free(eprecurse);

   return;
}


/*
 * Directory name comparison function, used by qsort
 */

static int
CompareDirEntry(const void *e1, const void *e2)
{
   struct DirEntry **ep1 = (struct DirEntry **) e1;
   struct DirEntry **ep2 = (struct DirEntry **) e2;

   return strcmp((*ep1)->name, (*ep2)->name);
}

/*
 * Change a directory.  Return a pointer to our new vdata structure for
 * this directory.
 */

static struct vnodeData *
ChangeDirectory(int argc, char **argv, struct vnodeData *vdatacwd)
{
   struct vnodeData *newvdatacwd;

   if (argc != 2) {
      fprintf(stderr, "Usage: %s directory\n", argv[0]);
      return NULL;
   }

   if ((newvdatacwd = FindFile(vdatacwd, argv[1])) == NULL)
      return NULL;

   if (newvdatacwd->vnode->type != vDirectory) {
      fprintf(stderr, "%s: Not a directory\n", argv[1]);
      return NULL;
   }

   if (newvdatacwd->filedata == NULL) {
      fprintf(stderr, "%s: No directory data found.\n", argv[1]);
      return NULL;
   }

   return newvdatacwd;
}

/*
 * Copy a file from out of the dump file
 */

#define COPYBUFSIZE 8192

static void
CopyFile(int argc, char **argv, struct vnodeData *vdatacwd, FILE *f)
{
   struct vnodeData *vdata;
   FILE *out;
   long cur = 0;
   int bytes, ret;
   char buffer[COPYBUFSIZE];

   if (argc != 3) {
      fprintf(stderr, "Usage: %s dumpfile destfile\n", argv[0]);
      return;
   }

   if ((vdata = FindFile(vdatacwd, argv[1])) == NULL)
      return;

   if (vdata->dumpdata == 0) {
      fprintf(stderr, "File %s has no data in dump file\n",
	      argv[1]);
      return;
   }

   if ((out = fopen(argv[2], "wb")) == NULL) {
      fprintf(stderr, "Open of %s failed: %s\n", argv[2],
	      strerror(errno));
      return;
   }

   if (fseek(f, vdata->dumpdata, SEEK_SET)) {
      fprintf(stderr, "Seek failed: %s\n", strerror(errno));
      fclose(out);
      return;
   }

   while (cur < vdata->datalength) {

      bytes = cur + COPYBUFSIZE < vdata->datalength ?
	 COPYBUFSIZE : vdata->datalength - cur;

      ret = fread(buffer, sizeof(char), bytes, f);
      if (ret != bytes) {
	 if (ret != 0)
	    fprintf(stderr, "Short read (expected %d, "
		    "got %d)\n", bytes, ret);
	 else
	    fprintf(stderr, "Error during read: %s\n",
		    strerror(errno));
	 fclose(out);
	 return;
      }

      ret = fwrite(buffer, sizeof(char), bytes, out);
      if (ret != bytes) {
	 if (ret != 0)
	    fprintf(stderr, "Short write (expected %d, "
		    "got %d)\n", bytes, ret);
	 else
	    fprintf(stderr, "Error during write: %s\n",
		    strerror(errno));
	 fclose(out);
	 return;
      }

      cur += bytes;
   }

   fclose(out);
}

/*
 * Copy a file from out of the dump file, by using the vnode
 */

static void
CopyVnode(int argc, char *argv[], FILE *f)
{
   struct vnodeData *vdata;
   FILE *out;
   long cur = 0;
   int bytes, ret;
   char buffer[COPYBUFSIZE];
   unsigned int vnode, uniquifier = 0;

   if (argc != 3) {
      fprintf(stderr, "Usage: %s vnode[.uniqifier] destfile\n",
	      argv[0]);
      return;
   }

   ret = sscanf(argv[1], "%d.%d", &vnode, &uniquifier);

   if (ret < 1) {
      fprintf(stderr, "Invalid file identifier: %s\n", argv[1]);
      return;
   }

   if (!(vdata = GetVnode(vnode))) {
      fprintf(stderr, "Vnode %d not in dump file\n", vnode);
      return;
   }

   if (ret == 2 && vdata->vnode->uniquifier != uniquifier) {
      fprintf(stderr, "Specified uniquifier %d did not match "
	      "uniquifier %d found in dump file!\n", uniquifier,
	      vdata->vnode->uniquifier);
      return;
   }

   if (vdata->dumpdata == 0) {
      fprintf(stderr, "File %s has no data in dump file\n",
	      argv[1]);
      return;
   }

   if ((out = fopen(argv[2], "wb")) == NULL) {
      fprintf(stderr, "Open of %s failed: %s\n", argv[2],
	      strerror(errno));
      return;
   }

   if (fseek(f, vdata->dumpdata, SEEK_SET)) {
      fprintf(stderr, "Seek failed: %s\n", strerror(errno));
      fclose(out);
      return;
   }

   while (cur < vdata->datalength) {

      bytes = cur + COPYBUFSIZE < vdata->datalength ?
	 COPYBUFSIZE : vdata->datalength - cur;

      ret = fread(buffer, sizeof(char), bytes, f);
      if (ret != bytes) {
	 if (ret != 0)
	    fprintf(stderr, "Short read (expected %d, "
		    "got %d)\n", bytes, ret);
	 else
	    fprintf(stderr, "Error during read: %s\n",
		    strerror(errno));
	 fclose(out);
	 return;
      }

      ret = fwrite(buffer, sizeof(char), bytes, out);
      if (ret != bytes) {
	 if (ret != 0)
	    fprintf(stderr, "Short write (expected %d, "
		    "got %d)\n", bytes, ret);
	 else
	    fprintf(stderr, "Error during write: %s\n",
		    strerror(errno));
	 fclose(out);
	 return;
      }

      cur += bytes;
   }

   fclose(out);
}
/*
 * Dump all residency filenames associated with a file, or all files
 * within a directory.
 */

static void
DumpAllFiles(int argc, char **argv, struct vnodeData *vdatacwd,
	     VolumeDiskData *vol)
{
   fprintf(stderr, "The \"file\" command is not available in the non-"
	   "MRAFS version of dumptool.\n");
   return;
}

/*
 * Given a directory vnode and a filename, return the vnode corresponding
 * to the file in that directory.
 * 
 * We now handle pathnames with directories in them.
 */

static struct vnodeData *
FindFile(struct vnodeData *vdatacwd, char *filename)
{
   struct DirHeader *dhp;
   struct DirEntry *ep;
   int i, num;
   struct vnodeData *vdata;
   char *c, newstr[MAXPATHLEN];

   if (! vdatacwd->filedata) {
      fprintf(stderr, "There is no vnode data for this "
	      "directory!\n");
      return NULL;
   }

   /*
    * If we have a "/" in here, look up the vnode data for the
    * directory (everything before the "/") and use that as our
    * current directory.  We automagically handle multiple directories
    * by using FindFile recursively.
    */

   if ((c = strrchr(filename, '/')) != NULL) {

      strncpy(newstr, filename, c - filename);
      newstr[c - filename] = '\0';

      if ((vdatacwd = FindFile(vdatacwd, newstr)) == NULL)
	 return NULL;

      if (vdatacwd->vnode->type != vDirectory) {
	 fprintf(stderr, "%s: Not a directory\n", newstr);
	 return NULL;
      }

      filename = c + 1;
   }

   dhp = (struct DirHeader *) vdatacwd->filedata;

   i = DirHash(filename);

   num = ntohs(dhp->hashTable[i]);

   while (num) {
      ep = (struct DirEntry *) (vdatacwd->filedata + (num * 32));
      if (strcmp(ep->name, filename) == 0)
	 break;
      num = ntohs(ep->next);
   }

   if (! num) {
      fprintf(stderr, "%s: No such file or directory\n", filename);
      return NULL;
   }

   if ((vdata = GetVnode(ntohl(ep->fid.vnode))) == NULL) {
      fprintf(stderr, "%s: No vnode information for %lu found\n",
	      filename, ntohl(ep->fid.vnode));
      return NULL;
   }

   return vdata;
}

/*
 * Reset a structure containing the current directory scan location
 */

static void
ResetDirCursor(struct DirCursor *cursor, struct vnodeData *vdata)
{
   struct DirHeader *dhp;

   cursor->hashbucket = 0;

   dhp = (struct DirHeader *) vdata->filedata;

   cursor->entry = ntohs(dhp->hashTable[0]);
}

/*
 * Given a cursor and a directory entry, return the next entry in the
 * directory.
 */

static struct DirEntry *
ReadNextDir(struct DirCursor *cursor, struct vnodeData *vdata)
{
   struct DirHeader *dhp;
   struct DirEntry *ep;

   dhp = (struct DirHeader *) vdata->filedata;

   if (cursor->entry) {
      ep = (struct DirEntry *) (vdata->filedata +
				(cursor->entry * 32));
      cursor->entry = ntohs(ep->next);
      return ep;
   } else {
      while (++(cursor->hashbucket) < NHASHENT) {
	 cursor->entry =
	    ntohs(dhp->hashTable[cursor->hashbucket]);
	 if (cursor->entry) {
	    ep = (struct DirEntry *) (vdata->filedata +
				      (cursor->entry * 32));
	    cursor->entry = ntohs(ep->next);
	    return ep;
	 }
      }
   }

   return NULL;
}

/*
 * Given a string, split it up into components a la Unix argc/argv.
 *
 * This code is most stolen from ftp.
 */

static void
MakeArgv(char *string, int *argc, char ***argv)
{
   static char *largv[64];
   char **la = largv;
   char *s = string;
   static char argbuf[256];
   char *ap = argbuf;

   *argc = 0;
   *argv = largv;

   while (*la++ = GetToken(s, &s, ap, &ap))
      (*argc)++;
}

/*
 * Return a pointer to the next token, and update the current string
 * position.
 */

static char *
GetToken(char *string, char **nexttoken, char argbuf[], char *nextargbuf[])
{
   char *sp = string;
   char *ap = argbuf;
   int got_one = 0;

  S0:
   switch (*sp) {

      case '\0':
	 goto OUTTOKEN;
	
      case ' ':
      case '\t':
	 sp++; goto S0;
	
      default:
	 goto S1;
   }

  S1:
   switch (*sp) {

      case ' ':
      case '\t':
      case '\0':
	 goto OUTTOKEN;	/* End of our token */

      case '\\':
	 sp++; goto S2;	/* Get next character */

      case '"':
	 sp++; goto S3;	/* Get quoted string */
	
      default:
	 *ap++ = *sp++;	/* Add a character to our token */
	 got_one = 1;
	 goto S1;
   }

  S2:
   switch (*sp) {

      case '\0':
	 goto OUTTOKEN;

      default:
	 *ap++ = *sp++;
	 got_one = 1;
	 goto S1;
   }

  S3:
   switch (*sp) {

      case '\0':
	 goto OUTTOKEN;
	
      case '"':
	 sp++; goto S1;
	
      default:
	 *ap++ = *sp++;
	 got_one = 1;
	 goto S3;
   }

  OUTTOKEN:
   if (got_one)
      *ap++ = '\0';
   *nextargbuf = ap;		/* Update storage pointer */
   *nexttoken = sp;		/* Update token pointer */

   return got_one ? argbuf : NULL;
}

/*
 * Insert vnodes into our hash table.
 */

static struct vnodeData *
InsertVnode(unsigned int vnodeNumber, struct VnodeDiskObject *vnode)
{
   struct VnodeDiskObject *nvnode;
   struct vnodeData *vdata;
   static int curSmallVnodeIndex = 0;
   static int curLargeVnodeIndex = 0;
   struct vnodeData ***vnodeIndex;
   int *curIndex;

   nvnode = (struct VnodeDiskObject *) malloc(sizeof(struct VnodeDiskObject));

   if (!nvnode) {
      if (verbose)
	 fprintf(stderr, "Unable to allocate space for vnode\n");
      return NULL;
   }

   memcpy((void *) nvnode, (void *) vnode, sizeof(struct VnodeDiskObject));

   if (vnodeNumber & 1) {
      vnodeIndex = &LargeVnodeIndex;
      curIndex = &curLargeVnodeIndex;
   } else {
      vnodeIndex = &SmallVnodeIndex;
      curIndex = &curSmallVnodeIndex;
   }

   vdata = (struct vnodeData *) malloc(sizeof(struct vnodeData));

   vdata->vnode = nvnode;
   vdata->vnodeNumber = vnodeNumber;
   vdata->dumpdata = 0;
   vdata->filedata = 0;
   vdata->datalength = 0;

   (*vnodeIndex)[(*curIndex)++] = vdata;

   return vdata;
}

/*
 * Routine to retrieve a vnode from the hash table.
 */

static struct vnodeData *
GetVnode(unsigned int vnodeNumber)
{
   struct vnodeData vnode, *vnodep, **tmp;

   vnode.vnodeNumber = vnodeNumber;
   vnodep = &vnode;

   tmp = (struct vnodeData **)
      bsearch((void *) &vnodep,
	      vnodeNumber & 1 ? LargeVnodeIndex : SmallVnodeIndex,
	      vnodeNumber & 1 ? numLargeVnodes : numSmallVnodes,
	      sizeof(struct vnodeData *), CompareVnode);

   return tmp ? *tmp : NULL;
}

/*
 * Our comparator function for bsearch
 */

static int
CompareVnode(const void *node1, const void *node2)
{
   struct vnodeData **vnode1 = (struct vnodeData **) node1;
   struct vnodeData **vnode2 = (struct vnodeData **) node2;

   if ((*vnode1)->vnodeNumber == (*vnode2)->vnodeNumber)
      return 0;
   else if ((*vnode1)->vnodeNumber > (*vnode2)->vnodeNumber)
      return 1;
   else
      return -1;
}

/*
 * Read a 16 bit integer in network order
 */

static int
ReadInt16(FILE *f, unsigned short *s)
{
   unsigned short in;

   if (fread((void *)&in, sizeof(in), 1, f) != 1) {
      if (verbose)
	 fprintf(stderr, "ReadInt16 failed!\n");
      return -1;
   }

   *s = ntohs(in);

   return 0;
}


/*
 * Read a 32 bit integer in network order
 */

static int
ReadInt32(FILE *f, unsigned int *i)
{
   unsigned int in;

   if (fread((void *)&in, sizeof(in), 1, f) != 1) {
      if (verbose)
	 fprintf(stderr, "ReadInt32 failed!\n");
      return -1;
   }

   *i = ntohl((unsigned long) in);

   return 0;
}

/*
 * Read a string from a dump file
 */

static int
ReadString(FILE *f, char *string, int maxlen)
{
   int c;

   while (maxlen--) {
      if ((*string++ = getc(f)) == 0)
	 break;
   }

   /*
    * I'm not sure what the _hell_ this is supposed to do ...
    * but it was in the original dump code
    */

   if (string[-1]) {
      while ((c = getc(f)) && c != EOF);
      string[-1] = 0;
   }

   return 0;
}

static int
ReadByteString(FILE *f, void *s, int size)
{
   unsigned char *c = (unsigned char *) s;

   while (size--)
      *c++ = getc(f);
	
   return 0;
}

/*
 * The directory hashing algorithm used by AFS
 */

DirHash (string)
   register char *string; {
   /* Hash a string to a number between 0 and NHASHENT. */
   register unsigned char tc;
   register int hval;
   register int tval;
   hval = 0;
   while(tc=(*string++)) {
      hval *= 173;
      hval  += tc;
   }
   tval = hval & (NHASHENT-1);
#ifdef AFS_CRAY_ENV     /* actually, any > 32 bit environment */
   if (tval == 0) return tval;
   else if (hval & 0x80000000) tval = NHASHENT-tval;
#else /* AFS_CRAY_ENV */
   if (tval == 0) return tval;
   else if (hval < 0) tval = NHASHENT-tval;
#endif /* AFS_CRAY_ENV */
   return tval;
}
