#include "volset.h"
#include "volset-link.h"
#include "volset-dump.h"
#include "volset-helper.h"
#include "volset-restore.h"
#include "volset-estimate.h"

/* This is how we will implement the volumesets:

   First, expand the volumeset, then cycle through the volume names.
   Write out the Dump Set Header.

   For each volume name:

   1. Spawn 'vos dump' on the volume (maybe get info on the volume before
      hand), with output going to us.

   2. Write out a chunk header

   3. Read a fixed block of the 'vos dump' output, escape all magic
      words found there.

   4. Write the block into the dump file.

   5. Continue until the sum of the read/written data equals chunk size.

   6. If there's more data in 'vos dump' output, proceed to the next chunk. 

   7. Otherwise, start new volume 'vos dump'

 */

int EvalVolumeSet(struct sockaddr_in server, afs_int32 partition, char *volname, Node **vols,
		  struct ubik_client *uclient)
{ /*EvalVolumeSet*/
   struct VldbListByAttributes attributes;
   afs_int32 si, nsi;            /* startIndex and nextStartIndex */
   afs_int32 nentries, e, ei, et, add, l, i;
   afs_int32 which_vol;
   nbulkentries bulkentries;
   struct nvldbentry *entries=0;
   afs_int32 code=0, tcode;
   afs_int32 count=0;
   struct vol *t;
   
   bulkentries.nbulkentries_len = 0;
   bulkentries.nbulkentries_val = 0;
   
   /* For the volume set entry - collect the volumes that match it */
   
   /* Put together a call to the vlserver for this vlentry. The 
    * performance gain is from letting the vlserver expand the
    * volumeset and not this routine.
    */
   attributes.Mask = 0;
   if (server.sin_addr.s_addr) {                       /* The server */
      attributes.Mask   |= VLLIST_SERVER;
      attributes.server  = server.sin_addr.s_addr;
   }
   if (partition != -1) {                              /* The partition */
      attributes.Mask      |= VLLIST_PARTITION;
      attributes.partition  = partition;
   }
   
   /* Now make the call to the vlserver */
   for (si=0; si != -1; si=nsi) {
      nentries = 0;
      bulkentries.nbulkentries_len = 0;
      bulkentries.nbulkentries_val = 0;
      nsi = -1;
      tcode = ubik_Call(VL_ListAttributesN2, uclient, 0, 
			&attributes, volname, si,
			&nentries, &bulkentries, &nsi);
      if (tcode) fprintf(stderr, "Error: Could not expand module lists!\n");

      if (nentries == 0)
      {
	 fprintf(stderr, "Error: nentries = 0!\n");
      }
      /* Step through each entry and add it to the list of volumes */
      entries = bulkentries.nbulkentries_val;
      for (e=0; e<nentries; e++) {
	 ei =  entries[e].matchindex        & 0xffff;
	 et = (entries[e].matchindex >> 16) & 0xffff;

/*	 switch (et) {
	    case ITSRWVOL:  {et = RWVOL;   break;}
	    case ITSBACKVOL:{et = BACKVOL; break;}
	    case ITSROVOL:  {et = ROVOL;   break;}
	    default:  { fprintf(stderr, "Error: strange volume type!\n"); }
	 }
*/ 

	 /* Allocate a volume dump structure and its name */
	 
	 t = (struct vol *) malloc(sizeof(struct vol));
	 
	 for (i=0; i<entries[e].nServers; i++) /* Find the RW volume */
	 {
	    if (entries[e].serverFlags[i] & et) break;
	 }
	 t->server = entries[e].serverNumber[i];
	 t->partition = entries[e].serverPartition[i];
	 t->volumeId[BACKVOL] = entries[e].volumeId[BACKVOL];
	 t->volumeId[ROVOL] = entries[e].volumeId[ROVOL];
	 t->volumeId[RWVOL] = entries[e].volumeId[RWVOL];
	 t->volname   = (char *) malloc(strlen(entries[e].name)+10);
	 if (!(t->volname)) {
	    fprintf(stderr, "Error: could not allocate memory!\n");
	 }
	 
	 /* Fill it in and thread onto avols list */
	 strcpy(t->volname, entries[e].name);
	 if      (et == ITSBACKVOL) strcat(t->volname, ".backup");
	 else if (et == ITSROVOL)   strcat(t->volname, ".readonly");

	 insert(vols, t);

	 count++;
      }
      
      /* Free memory allocated during VL call */
      if (bulkentries.nbulkentries_val) {
	 free((char *)bulkentries.nbulkentries_val);
	 bulkentries.nbulkentries_val = 0;
	 entries = 0;
      } 
   }

   /* Randomly link the volumedump entries together */
/*   fprintf (stderr, "Total number of volumes : %u\n", count); */
   
  error_exit:
   if (bulkentries.nbulkentries_val) {
      free((char *)bulkentries.nbulkentries_val);
   }
   return(code);
} /*EvalVolumeSet*/

int main(int argc, char *argv[])
{
   struct ubik_client *cstruct;
   char cmd[32];
   char part[32];
   afs_int32 partition;
   char volname[32];
   Node *vols;
   char serv[32];
   char file[32];
   struct sockaddr_in server;
   afs_int32 code;
   afs_int32 i;
   afs_int32 sauth;
   char *tcell;
   int rxpackets = 100;
   int out = -1;
   FILE *fout;
   time_t prev_time;
   EIGHT_BYTE_INT written = 0;
   double dwritten = 0;
   extern int UV_SetSecurity();

   vols = NULL;
   if(argv[1] != '\000')
   {
      sscanf(argv[1], "%s", cmd);
   }
   else
   {
      fprintf(stderr, "%s\n", AMANDA_AFS_VERSION);
      fprintf(stderr, "Usage:\tvolset <cmd> <cmd-options>\n", argv[0]);
      fprintf(stderr, "\tWhere <cmd> is 'dump', 'restore', or 'estimate'\n\n");
      fprintf(stderr, "\tDump options: <server> <partition> <volume pattern> <dump time> <dump file>\n");
      fprintf(stderr, "\tRestore options: <dump file> <volume to restore>|* <restore file>|afs:<server>/<partition>\n");
      fprintf(stderr, "\tEstimate options: <server> <partition> <volume pattern> <dump time>\n\n");
      fprintf(stderr, "\tIf * is specified, then <restore file> should be a directory\n");
      fprintf(stderr, "\totherwise it's a filename.\n");
      fprintf(stderr, "\tIf afs:<server>/<partition> is specified, then the volumes specified to restore will be\n"
	      "\trestored to the specified partition.\n\n");
      fprintf(stderr, "\tAll options are MANDATORY!\n");
      return 0;
   }

   if(strcmp(cmd, "dump") == 0 || strcmp(cmd, "estimate") == 0)
   {
      if(argc < 6)
      {
	 fprintf(stderr, "Not enough cmdline arguments!\nTo get help, run this command without any arguments.\n", argv[0]);
	 return 1;
      }
      sscanf(argv[2], "%s", serv);
      sscanf(argv[3], "%s", part);
      sscanf(argv[4], "%s", volname);
      sscanf(argv[5], "%ld", &prev_time);

      code = ParseHost(serv, &server);
      code = GetPartitionID(&partition, part);
         
      VInitVolumePackage(volumeUtility, 0, 0, CONNECT_FS, 0); 
/*   DInit(40); */
      
      /* Initialize the ubik_client connection */
      rx_SetRxDeadTime(90);
      cstruct = (struct ubik_client *)0;
      
      sauth = 0;
      tcell = (char *) 0;
      if (code = vsu_ClientInit(1, AFSDIR_CLIENT_ETC_DIRPATH, tcell,
				sauth, &cstruct, UV_SetSecurity)) {
	 fprintf(STDERR,"could not initialize VLDB library (code=%u) \n",code);
	 exit(1);
      }
      
      /* code here */
      
      code = EvalVolumeSet(server, partition, volname, &vols, cstruct);
      i=0;

      if((argc > 6) && (strcmp(cmd, "dump") == 0))
      {
	 sscanf(argv[6], "%s", file);

	 if(strcmp(file, "-") == 0)
	 {
	    out = fileno(stdout);
	 }
	 else
	 {
	    out = open(file, O_WRONLY | O_TRUNC , S_IRUSR | S_IWUSR);
	    if(out == -1)
	    {
	       out = open(file, O_CREAT | O_WRONLY | O_TRUNC, S_IRUSR | S_IWUSR);
	       if(out == -1)
	       {
		  fprintf(stderr, "Error: %d!\n", errno);
	       }
	    }
	 }
	    
	 if(DumpVolSetHeader(out, &vols))
	 {
	    fprintf(stderr, "Error in DumpVolSetHeader!\n");
	 }
	 
	 while(vols)
	 {
	    written += DumpVolume(out, get(&vols), prev_time, vols->next?0:1);
	 }
	 
	 close(out);
	 
      }
      else if((argc <= 6) && (strcmp(cmd, "dump") == 0))
      {
	 fprintf(stderr, "Not enough cmdline arguments!\nTo get help, run this command without any arguments.\n", argv[0]);
	 if (cstruct) {
	    ubik_ClientDestroy (cstruct);
	    cstruct = 0;
	 }
	 rx_Finalize();
	 return 1;
      }
      else if(strcmp(cmd, "estimate") == 0)
      {
	 while(vols)
	 {
	    written += EstimateVolume(get(&vols), prev_time);
	 }
      }
      
      dwritten = written/1024.0/1024.0;
      fprintf(stderr, "Total bytes written: ");
      fprintf(stderr, "%lld", written);
      fprintf(stderr, " (%.2lfMB, ?MB/s) \n", dwritten);
      /* end code here */
      
      if (cstruct) {
	 ubik_ClientDestroy (cstruct);
	 cstruct = 0;
      }
      rx_Finalize();
   }
   else if(strcmp(cmd, "restore") == 0)
   {
      if(argc < 5)
      {
	 fprintf(stderr, "Not enough cmdline arguments!\nTo get help, run this command without any arguments.\n", argv[0]);
	 return 1;
      }
      sscanf(argv[2], "%s", file);
      sscanf(argv[3], "%s", volname);
      sscanf(argv[4], "%s", part);

      if(strcmp(file, "-") == 0)
      {
	 fout = stdin;
      }
      else
      {
	 fout = fopen(file, "r");
      }
      
      if(fout == NULL)
      {
	 fprintf(stderr, "Error: could not open file: %d!\n", errno);
	 return 0;
      }
      
      if(ReadVolSetHeader(fout, &vols))
      {
	 fprintf(stderr, "Error in ReadVolSetHeader!\n");
      }
      
      if(RestoreVolume(fout, part, volname) == -1)
      {
	 fprintf(stderr, "Error in RestoreVolume!\n");
      }
      
      fclose(fout);
   }
   
   return 0;
}
